/*
Copyright 2011 Google Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// Package memcache provides a client for the memcached cache server.
package memcache

import (
	"bufio"
	"bytes"
	"errors"
	"fmt"
	"io"
	"net"

	"context"
	"strconv"
	"strings"
	"sync"
	"sync/atomic"
	"time"
)

// Similar to:
// http://code.google.com/appengine/docs/go/memcache/reference.html

var (
	// ErrCacheMiss means that a Get failed because the item wasn't present.
	ErrCacheMiss = errors.New("memcache: cache miss")

	// ErrCASConflict means that a CompareAndSwap call failed due to the
	// cached value being modified between the Get and the CompareAndSwap.
	// If the cached value was simply evicted rather than replaced,
	// ErrNotStored will be returned instead.
	ErrCASConflict = errors.New("memcache: compare-and-swap conflict")

	// ErrNotStored means that a conditional write operation (i.e. Add or
	// CompareAndSwap) failed because the condition was not satisfied.
	ErrNotStored = errors.New("memcache: item not stored")

	// ErrServerError means that a server error occurred.
	ErrServerError = errors.New("memcache: server error")

	// ErrNoStats means that no statistics were available.
	ErrNoStats = errors.New("memcache: no statistics available")

	// ErrMalformedKey is returned when an invalid key is used.
	// Keys must be at maximum 250 bytes long and not
	// contain whitespace or control characters.
	ErrMalformedKey = errors.New("malformed: key is too long or contains invalid characters")

	// ErrNoServers is returned when no servers are configured or available.
	ErrNoServers = errors.New("memcache: no servers configured or available")
)

const (
	// DefaultTimeout is the default socket read/write timeout.
	DefaultTimeout = 100 * time.Millisecond

	// DefaultMaxIdleConns is the default maximum number of idle connections
	// kept for any single address.
	DefaultMaxIdleConns = 2
)

const buffered = 8 // arbitrary buffered channel size, for readability

// resumableError returns true if err is only a protocol-level cache error.
// This is used to determine whether or not a server connection should
// be re-used or not. If an error occurs, by default we don't reuse the
// connection, unless it was just a cache error.
func resumableError(err error) bool {
	switch err {
	case ErrCacheMiss, ErrCASConflict, ErrNotStored, ErrMalformedKey:
		return true
	}
	return false
}

func legalKey(key string) bool {
	if len(key) > 250 {
		return false
	}
	for i := 0; i < len(key); i++ {
		if key[i] <= ' ' || key[i] == 0x7f {
			return false
		}
	}
	return true
}

var (
	crlf            = []byte("\r\n")
	space           = []byte(" ")
	resultOK        = []byte("OK\r\n")
	resultStored    = []byte("STORED\r\n")
	resultNotStored = []byte("NOT_STORED\r\n")
	resultExists    = []byte("EXISTS\r\n")
	resultNotFound  = []byte("NOT_FOUND\r\n")
	resultDeleted   = []byte("DELETED\r\n")
	resultEnd       = []byte("END\r\n")
	resultTouched   = []byte("TOUCHED\r\n")

	resultClientErrorPrefix = []byte("CLIENT_ERROR ")
)

// New returns a memcache client using the provided server(s)
// with equal weight. If a server is listed multiple times,
// it gets a proportional amount of weight.
func New(server ...string) (*Client, error) {
	ss := new(ServerList)
	if err := ss.SetServers(server...); err != nil {
		return nil, err
	}
	return NewFromSelector(ss), nil
}

// NewFromSelector returns a new Client using the provided ServerSelector.
func NewFromSelector(ss ServerSelector) *Client {
	return &Client{selector: ss}
}

// Client is a memcache client.
// It is safe for unlocked use by multiple concurrent goroutines.
type Client struct {
	// Timeout specifies the socket read/write timeout.
	// If zero, DefaultTimeout is used.
	Timeout time.Duration
	Dialer  net.Dialer
	// DialFunc specifies a custom 'DialContext' function.
	// Callers can use this to route connections through a proxy or alternative network type.
	DialFunc func(ctx context.Context, network, address string) (net.Conn, error)

	// MaxIdleConns specifies the maximum number of idle connections that will
	// be maintained per address. If less than one, DefaultMaxIdleConns will be
	// used.
	//
	// Consider your expected traffic rates and latency carefully. This should
	// be set to a number higher than your peak parallel requests.
	MaxIdleConns int

	selector ServerSelector

	lk       sync.Mutex
	freeconn map[string][]*conn

	stats Stats
}

// Stats allows memcache client users to collect information about the current state of the client
type Stats struct {
	// Timeout is the client's configured timeout
	Timeout time.Duration
	// MaxIdleConns is the client's configured maximum idle connections
	MaxIdleConns int
	// Number of failed close calls.
	FailedClose int64
	// When a MC call can reuse a connection, rather than make a new one
	ReusedConnections int64
	// When a MC connection is sent to the pool
	ReturnedConnections int64
	// When a MC connection is created
	CreatedConnections int64
	AddressStats       map[string]struct {
		PoolSize int
	}
	// The number of times an idempotent request was retried due to a client failure
	IdempotentRetries int64
}

// ClientStats returns an instance in time snapshot of the client's current state
// This is *not* the memcache `stats`, which would be server state.
func (c *Client) ClientStats() Stats {
	ret := Stats{
		Timeout:             c.Timeout,
		MaxIdleConns:        c.MaxIdleConns,
		FailedClose:         atomic.LoadInt64(&c.stats.FailedClose),
		ReusedConnections:   atomic.LoadInt64(&c.stats.ReusedConnections),
		ReturnedConnections: atomic.LoadInt64(&c.stats.ReturnedConnections),
		CreatedConnections:  atomic.LoadInt64(&c.stats.CreatedConnections),
		IdempotentRetries:   atomic.LoadInt64(&c.stats.IdempotentRetries),
	}
	c.lk.Lock()
	defer c.lk.Unlock()
	ret.AddressStats = make(map[string]struct{ PoolSize int }, len(c.freeconn))
	for addr, conns := range c.freeconn {
		ret.AddressStats[addr] = struct {
			PoolSize int
		}{
			PoolSize: len(conns),
		}
	}
	return ret
}

// Item is an item to be got or stored in a memcached server.
type Item struct {
	// Key is the Item's key (250 bytes maximum).
	Key string

	// Value is the Item's value.
	Value []byte

	// Flags are server-opaque flags whose semantics are entirely
	// up to the app.
	Flags uint32

	// Expiration is the cache expiration time, in seconds: either a relative
	// time from now (up to 1 month), or an absolute Unix epoch time.
	// Zero means the Item has no expiration time.
	Expiration int32

	// Compare and swap ID.
	casid uint64
}

// conn is a connection to a server.
type conn struct {
	nc   net.Conn
	rw   *bufio.ReadWriter
	addr net.Addr
	c    *Client
	// used is true if the connection has been used before.  For idempotent
	// requests, we retry them only if the connection was previously used.  This
	// case is common for memcache, where idle connections hang around and get dropped
	// when the client eventually tries to use them.
	used bool
}

// release returns this connection back to the client's free pool
func (cn *conn) release() {
	cn.c.putFreeConn(cn.addr, cn)
}

// Extend connections deadline using the context.  Follows this resolution order:
//   1) If ctx.Deadline() exists, use the context's deadline
//   2) If Timeout is set, use time.Now().Add(timeout)
//   3) Use time.Now().Add(DefaultTimeout)
func (cn *conn) extendDeadline(ctx context.Context) error {
	deadline, exists := ctx.Deadline()
	if exists {
		return cn.nc.SetDeadline(deadline)
	}
	if cn.c.Timeout != 0 {
		deadline = time.Now().Add(cn.c.Timeout)
	} else {
		deadline = time.Now().Add(DefaultTimeout)
	}
	return cn.nc.SetDeadline(deadline)
}

// condRelease releases this connection if the error pointed to by err
// is nil (not an error) or is only a protocol level error (e.g. a
// cache miss).  The purpose is to not recycle TCP connections that
// are bad.
// Note: We removed the *error value since we're no longer using this inside
//       defer.
// Returns true if the error was not a signal that the socket connection was bad.
func (cn *conn) condRelease(err error) bool {
	if err == nil || resumableError(err) {
		cn.release()
		return true
	}
	if err := cn.nc.Close(); err != nil {
		atomic.AddInt64(&cn.c.stats.FailedClose, 1)
	}
	return false
}

func (c *Client) putFreeConn(addr net.Addr, cn *conn) { // nolint: interfacer
	c.lk.Lock()
	defer c.lk.Unlock()
	// This connection is 'used'.  That means it's been sitting in a pool for a while.
	// If a request fails because of this, it's possible the failure was because the
	// connection was no longer good.
	cn.used = true
	if c.freeconn == nil {
		c.freeconn = make(map[string][]*conn)
	}
	freelist := c.freeconn[addr.String()]
	if len(freelist) >= c.maxIdleConns() {
		if err := cn.nc.Close(); err != nil {
			atomic.AddInt64(&c.stats.FailedClose, 1)
		}
		return
	}
	atomic.AddInt64(&c.stats.ReturnedConnections, 1)
	c.freeconn[addr.String()] = append(freelist, cn)
}

func (c *Client) getFreeConn(addr net.Addr) (cn *conn, ok bool) { // nolint: interfacer
	c.lk.Lock()
	defer c.lk.Unlock()
	if c.freeconn == nil {
		return nil, false
	}
	freelist, ok := c.freeconn[addr.String()]
	if !ok || len(freelist) == 0 {
		return nil, false
	}
	cn = freelist[len(freelist)-1]
	c.freeconn[addr.String()] = freelist[:len(freelist)-1]
	atomic.AddInt64(&c.stats.ReusedConnections, 1)
	return cn, true
}

func (c *Client) maxIdleConns() int {
	if c.MaxIdleConns > 0 {
		return c.MaxIdleConns
	}
	return DefaultMaxIdleConns
}

// ConnectTimeoutError is the error type used when it takes
// too long to connect to the desired host. This level of
// detail can generally be ignored.
type ConnectTimeoutError struct {
	Addr net.Addr
}

func (cte *ConnectTimeoutError) Error() string {
	return "memcache: connect timeout to " + cte.Addr.String()
}

func (c *Client) dial(ctx context.Context, addr net.Addr) (net.Conn, error) {
	atomic.AddInt64(&c.stats.CreatedConnections, 1)

	var nc net.Conn
	var err error
	if c.DialFunc != nil {
		nc, err = c.DialFunc(ctx, addr.Network(), addr.String())
	} else {
		nc, err = c.Dialer.DialContext(ctx, addr.Network(), addr.String())
	}
	if err == nil {
		return nc, nil
	}

	if ne, ok := err.(net.Error); ok && ne.Timeout() {
		return nil, &ConnectTimeoutError{addr}
	}

	return nil, err
}

// getConn returns a connection for addr, possibly going to the pool.  If forceFresh is true,
// then it will skip checking the pool and always return a fresh connection.
func (c *Client) getConn(ctx context.Context, addr net.Addr, forceFresh bool) (*conn, error) {
	if !forceFresh {
		cn, ok := c.getFreeConn(addr)
		if ok {
			err := cn.extendDeadline(ctx)
			if err == nil {
				return cn, nil
			}
			cn.condRelease(err)
		}
	}
	// If we forced a fresh connection, it was because we are retrying an idempotent request
	if forceFresh {
		atomic.AddInt64(&c.stats.IdempotentRetries, 1)
	}
	nc, err := c.dial(ctx, addr)
	if err != nil {
		return nil, err
	}
	cn := &conn{
		nc:   nc,
		addr: addr,
		rw:   bufio.NewReadWriter(bufio.NewReader(nc), bufio.NewWriter(nc)),
		c:    c,
	}
	return cn, cn.extendDeadline(ctx)
}

func (c *Client) onItem(ctx context.Context, item *Item, fn func(*Client, *bufio.ReadWriter, *Item) error, idempotent bool) error {
	addr, err := c.selector.PickServer(item.Key)
	if err != nil {
		return err
	}
	cn, err := c.getConn(ctx, addr, false)
	if err != nil {
		return err
	}
	err = fn(c, cn.rw, item)
	if cn.condRelease(err) || !cn.used {
		return err
	}
	// We couldn't release this connection back to the pool.  This was
	// probably because the error was network related. If the request
	// is idempotent, let's try again on a forced fresh connection
	cn, err = c.getConn(ctx, addr, true)
	if err != nil {
		return err
	}
	err = fn(c, cn.rw, item)
	cn.condRelease(err)
	return err
}

// FlushAll sends the flush_all command
func (c *Client) FlushAll(ctx context.Context) error {
	return c.selector.Each(ctx, c.flushAllFromAddr)
}

// Get gets the item for the given key. ErrCacheMiss is returned for a
// memcache cache miss. The key must be at most 250 bytes in length.
func (c *Client) Get(ctx context.Context, key string) (item *Item, err error) {
	err = c.withKeyAddr(key, func(addr net.Addr) error {
		return c.getFromAddr(ctx, addr, []string{key}, func(it *Item) { item = it })
	})
	if err == nil && item == nil {
		err = ErrCacheMiss
	}
	return
}

// Touch updates the expiry for the given key. The seconds parameter is either
// a Unix timestamp or, if seconds is less than 1 month, the number of seconds
// into the future at which time the item will expire. ErrCacheMiss is returned if the
// key is not in the cache. The key must be at most 250 bytes in length.
func (c *Client) Touch(ctx context.Context, key string, seconds int32) (err error) {
	return c.withKeyAddr(key, func(addr net.Addr) error {
		return c.touchFromAddr(ctx, addr, []string{key}, seconds)
	})
}

func (c *Client) withKeyAddr(key string, fn func(net.Addr) error) error {
	if !legalKey(key) {
		return ErrMalformedKey
	}
	addr, err := c.selector.PickServer(key)
	if err != nil {
		return err
	}
	return fn(addr)
}

func (c *Client) withAddrRw(ctx context.Context, addr net.Addr, fn func(*bufio.ReadWriter) error, idempotent bool) error {
	cn, err := c.getConn(ctx, addr, false)
	if err != nil {
		return err
	}
	err = fn(cn.rw)
	if cn.condRelease(err) || !cn.used {
		return err
	}
	// This result was probably a networking error, the connection was somewhat stale,
	// and the request was idempotent.  Try it again on a forced fresh connection.
	cn, err = c.getConn(ctx, addr, true)
	if err != nil {
		return err
	}
	err = fn(cn.rw)
	cn.condRelease(err)
	return err
}

func (c *Client) withKeyRw(ctx context.Context, key string, fn func(*bufio.ReadWriter) error, idempotent bool) error {
	return c.withKeyAddr(key, func(addr net.Addr) error {
		return c.withAddrRw(ctx, addr, fn, idempotent)
	})
}

func (c *Client) getFromAddr(ctx context.Context, addr net.Addr, keys []string, cb func(*Item)) error {
	return c.withAddrRw(ctx, addr, func(rw *bufio.ReadWriter) error {
		if _, err := fmt.Fprintf(rw, "gets %s\r\n", strings.Join(keys, " ")); err != nil {
			return err
		}
		if err := rw.Flush(); err != nil {
			return err
		}
		if err := parseGetResponse(rw.Reader, cb); err != nil {
			return err
		}
		return nil
	}, true)
}

// flushAllFromAddr send the flush_all command to the given addr
func (c *Client) flushAllFromAddr(ctx context.Context, addr net.Addr) error {
	return c.withAddrRw(ctx, addr, func(rw *bufio.ReadWriter) error {
		if _, err := fmt.Fprintf(rw, "flush_all\r\n"); err != nil {
			return err
		}
		if err := rw.Flush(); err != nil {
			return err
		}
		line, err := rw.ReadSlice('\n')
		if err != nil {
			return err
		}
		switch {
		case bytes.Equal(line, resultOK):
			break
		default:
			return fmt.Errorf("memcache: unexpected response line from flush_all: %q", string(line))
		}
		return nil
	}, false)
}

func (c *Client) touchFromAddr(ctx context.Context, addr net.Addr, keys []string, expiration int32) error {
	return c.withAddrRw(ctx, addr, func(rw *bufio.ReadWriter) error {
		for _, key := range keys {
			if _, err := fmt.Fprintf(rw, "touch %s %d\r\n", key, expiration); err != nil {
				return err
			}
			if err := rw.Flush(); err != nil {
				return err
			}
			line, err := rw.ReadSlice('\n')
			if err != nil {
				return err
			}
			switch {
			case bytes.Equal(line, resultTouched):
				// Good case
			case bytes.Equal(line, resultNotFound):
				return ErrCacheMiss
			default:
				return fmt.Errorf("memcache: unexpected response line from touch: %q", string(line))
			}
		}
		return nil
	}, false)
}

// GetMulti is a batch version of Get. The returned map from keys to
// items may have fewer elements than the input slice, due to memcache
// cache misses. Each key must be at most 250 bytes in length.
// If no error is returned, the returned map will also be non-nil.
func (c *Client) GetMulti(ctx context.Context, keys []string) (map[string]*Item, error) {
	var lk sync.Mutex
	m := make(map[string]*Item)
	addItemToMap := func(it *Item) {
		lk.Lock()
		defer lk.Unlock()
		m[it.Key] = it
	}

	keyMap := make(map[net.Addr][]string)
	for _, key := range keys {
		if !legalKey(key) {
			return nil, ErrMalformedKey
		}
		addr, err := c.selector.PickServer(key)
		if err != nil {
			return nil, err
		}
		keyMap[addr] = append(keyMap[addr], key)
	}

	ch := make(chan error, buffered)
	for addr, keys := range keyMap {
		go func(addr net.Addr, keys []string) {
			ch <- c.getFromAddr(ctx, addr, keys, addItemToMap)
		}(addr, keys)
	}

	var err error
	for range keyMap {
		if ge := <-ch; ge != nil {
			err = ge
		}
	}
	return m, err
}

// parseGetResponse reads a GET response from r and calls cb for each
// read and allocated Item
func parseGetResponse(r *bufio.Reader, cb func(*Item)) error {
	for {
		line, err := r.ReadSlice('\n')
		if err != nil {
			return err
		}
		if bytes.Equal(line, resultEnd) {
			return nil
		}
		it := new(Item)
		size, err := scanGetResponseLine(line, it)
		if err != nil {
			return err
		}
		it.Value = make([]byte, size+2)
		_, err = io.ReadFull(r, it.Value)
		if err != nil {
			it.Value = nil
			return err
		}
		if !bytes.HasSuffix(it.Value, crlf) {
			it.Value = nil
			return fmt.Errorf("memcache: corrupt get result read")
		}
		it.Value = it.Value[:size]
		cb(it)
	}
}

// scanGetResponseLine populates it and returns the declared size of the item.
// It does not read the bytes of the item.
func scanGetResponseLine(line []byte, it *Item) (size int, err error) {
	pattern := "VALUE %s %d %d %d\r\n"
	dest := []interface{}{&it.Key, &it.Flags, &size, &it.casid}
	if bytes.Count(line, space) == 3 {
		pattern = "VALUE %s %d %d\r\n"
		dest = dest[:3]
	}
	n, err := fmt.Sscanf(string(line), pattern, dest...)
	if err != nil || n != len(dest) {
		return -1, fmt.Errorf("memcache: unexpected line in get response: %q", line)
	}
	return size, nil
}

// Set writes the given item, unconditionally.
func (c *Client) Set(ctx context.Context, item *Item) error {
	return c.onItem(ctx, item, (*Client).set, true)
}

func (c *Client) set(rw *bufio.ReadWriter, item *Item) error {
	return c.populateOne(rw, "set", item)
}

// Add writes the given item, if no value already exists for its
// key. ErrNotStored is returned if that condition is not met.
func (c *Client) Add(ctx context.Context, item *Item) error {
	return c.onItem(ctx, item, (*Client).add, false)
}

func (c *Client) add(rw *bufio.ReadWriter, item *Item) error {
	return c.populateOne(rw, "add", item)
}

// Replace writes the given item, but only if the server *does*
// already hold data for this key
func (c *Client) Replace(ctx context.Context, item *Item) error {
	return c.onItem(ctx, item, (*Client).replace, false)
}

func (c *Client) replace(rw *bufio.ReadWriter, item *Item) error {
	return c.populateOne(rw, "replace", item)
}

// CompareAndSwap writes the given item that was previously returned
// by Get, if the value was neither modified or evicted between the
// Get and the CompareAndSwap calls. The item's Key should not change
// between calls but all other item fields may differ. ErrCASConflict
// is returned if the value was modified in between the
// calls. ErrNotStored is returned if the value was evicted in between
// the calls.
func (c *Client) CompareAndSwap(ctx context.Context, item *Item) error {
	return c.onItem(ctx, item, (*Client).cas, false)
}

func (c *Client) cas(rw *bufio.ReadWriter, item *Item) error {
	return c.populateOne(rw, "cas", item)
}

func (c *Client) populateOne(rw *bufio.ReadWriter, verb string, item *Item) error {
	if !legalKey(item.Key) {
		return ErrMalformedKey
	}
	var err error
	if verb == "cas" {
		_, err = fmt.Fprintf(rw, "%s %s %d %d %d %d\r\n",
			verb, item.Key, item.Flags, item.Expiration, len(item.Value), item.casid)
	} else {
		_, err = fmt.Fprintf(rw, "%s %s %d %d %d\r\n",
			verb, item.Key, item.Flags, item.Expiration, len(item.Value))
	}
	if err != nil {
		return err
	}
	if _, err = rw.Write(item.Value); err != nil {
		return err
	}
	if _, err = rw.Write(crlf); err != nil {
		return err
	}
	if err = rw.Flush(); err != nil {
		return err
	}
	line, err := rw.ReadSlice('\n')
	if err != nil {
		return err
	}
	switch {
	case bytes.Equal(line, resultStored):
		return nil
	case bytes.Equal(line, resultNotStored):
		return ErrNotStored
	case bytes.Equal(line, resultExists):
		return ErrCASConflict
	case bytes.Equal(line, resultNotFound):
		return ErrCacheMiss
	}
	return fmt.Errorf("memcache: unexpected response line from %q: %q", verb, string(line))
}

func writeReadLine(rw *bufio.ReadWriter, format string, args ...interface{}) ([]byte, error) {
	_, err := fmt.Fprintf(rw, format, args...)
	if err != nil {
		return nil, err
	}
	if err = rw.Flush(); err != nil {
		return nil, err
	}
	line, err := rw.ReadSlice('\n')
	return line, err
}

func writeExpectf(rw *bufio.ReadWriter, expect []byte, format string, args ...interface{}) error {
	line, err := writeReadLine(rw, format, args...)
	if err != nil {
		return err
	}
	switch {
	case bytes.Equal(line, resultOK):
		return nil
	case bytes.Equal(line, expect):
		return nil
	case bytes.Equal(line, resultNotStored):
		return ErrNotStored
	case bytes.Equal(line, resultExists):
		return ErrCASConflict
	case bytes.Equal(line, resultNotFound):
		return ErrCacheMiss
	}
	return fmt.Errorf("memcache: unexpected response line: %q", string(line))
}

// Delete deletes the item with the provided key. The error ErrCacheMiss is
// returned if the item didn't already exist in the cache.
func (c *Client) Delete(ctx context.Context, key string) error {
	return c.withKeyRw(ctx, key, func(rw *bufio.ReadWriter) error {
		return writeExpectf(rw, resultDeleted, "delete %s\r\n", key)
	}, false)
}

// IdempotentDelete will retry the delete as if it was idempotent: allowing
// you to recover from stale client connections.  It is possible that the
// returned value, ErrCacheMiss, could happen if the second, retried request
// couldn't find an item that the first request was able to remove.
func (c *Client) IdempotentDelete(ctx context.Context, key string) error {
	return c.withKeyRw(ctx, key, func(rw *bufio.ReadWriter) error {
		return writeExpectf(rw, resultDeleted, "delete %s\r\n", key)
	}, true)
}

// DeleteAll deletes all items in the cache.
func (c *Client) DeleteAll(ctx context.Context) error {
	return c.withKeyRw(ctx, "", func(rw *bufio.ReadWriter) error {
		return writeExpectf(rw, resultDeleted, "flush_all\r\n")
	}, false)
}

// Increment atomically increments key by delta. The return value is
// the new value after being incremented or an error. If the value
// didn't exist in memcached the error is ErrCacheMiss. The value in
// memcached must be an decimal number, or an error will be returned.
// On 64-bit overflow, the new value wraps around.
func (c *Client) Increment(ctx context.Context, key string, delta uint64) (newValue uint64, err error) {
	return c.incrDecr(ctx, "incr", key, delta)
}

// Decrement atomically decrements key by delta. The return value is
// the new value after being decremented or an error. If the value
// didn't exist in memcached the error is ErrCacheMiss. The value in
// memcached must be an decimal number, or an error will be returned.
// On underflow, the new value is capped at zero and does not wrap
// around.
func (c *Client) Decrement(ctx context.Context, key string, delta uint64) (newValue uint64, err error) {
	return c.incrDecr(ctx, "decr", key, delta)
}

func (c *Client) incrDecr(ctx context.Context, verb, key string, delta uint64) (uint64, error) {
	var val uint64
	err := c.withKeyRw(ctx, key, func(rw *bufio.ReadWriter) error {
		line, err := writeReadLine(rw, "%s %s %d\r\n", verb, key, delta)
		if err != nil {
			return err
		}
		switch {
		case bytes.Equal(line, resultNotFound):
			return ErrCacheMiss
		case bytes.HasPrefix(line, resultClientErrorPrefix):
			errMsg := line[len(resultClientErrorPrefix) : len(line)-2]
			return errors.New("memcache: client error: " + string(errMsg))
		}
		val, err = strconv.ParseUint(string(line[:len(line)-2]), 10, 64)
		return err
	}, false)
	return val, err
}
